  <nav class="navbar navbar-expand-lg navbar-dark bg-secondary mb-5">
    <button class="navbar-toggler" type="button" data-toggle="collapse" data-target="#navbarSupportedContent" aria-controls="navbarSupportedContent" aria-expanded="false" aria-label="Toggle navigation">
      <span class="navbar-toggler-icon"></span>
    </button>

      <div class="collapse navbar-collapse" id="navbarSupportedContent">
        <ul class="navbar-nav mr-auto">
          {% for item, url in menu.items %}
          <li class="nav-item">
            <a class="nav-link {% if item == active %}active{% endif %}" href="{{url}}">
              {{item}}
              {% if item == active %}<span class="sr-only">(current)</span>{% endif %}
              </a>
          </li>
          {% endfor %}
        </ul>
      </div>

  </nav>
